' LnknLogs.Bas - Linked list manipulation package

' $INCLUDE: 'LNKNTYPS.BI'
' $INCLUDE: 'LNKNLOGS.BI'
' $INCLUDE: 'LNKNPRVT.BI'

SUB NewList (ListHandle AS INTEGER)
' NewList -   Allocate a new linked list and return a handle to it
'		in the argument ListHandle.
DIM WorkHdl AS INTEGER
ListHandle = nil
WorkHdl = 1
DO WHILE WorkHdl < MaxList + 1 AND ListHandle = nil
  IF ListHeads(WorkHdl) = nil THEN
    ListHandle = WorkHdl
  ELSE
    WorkHdl = WorkHdl + 1
  END IF
LOOP
END SUB

SUB DisposeList (ListHandle AS INTEGER)
' DisposeList - De-allocate the list specified by ListHandle
'		and mark as reusable any nodes linked into it.
DIM WorkPtr AS INTEGER
DIM NextPtr AS INTEGER
IF ListHandle <> nil THEN
  WorkPtr = ListHeads(ListHandle)
  DO WHILE WorkPtr <> nil
    NextPtr = NodeSpace(WorkPtr).fLink
    MarkNode WorkPtr
    WorkPtr = NextPtr
  LOOP
  ListHeads(ListHandle) = nil
  ListHandle = nil
END IF
END SUB

SUB NewNode (NodePtr AS INTEGER)
' NewNode - Allocate a new list node, growing the allocation
'   space if necessary
DIM WorkPtr AS INTEGER
NodePtr = nil
WorkPtr = 1
DO WHILE WorkPtr < MaxNode + 1 AND NodePtr = nil
  IF NodeSpace(WorkPtr).fLink = unused THEN
    NodePtr = WorkPtr
  ELSE
    WorkPtr = WorkPtr + 1
  END IF
LOOP
IF NodePtr = nil THEN
  GrowAllocZone
  IF UBOUND(NodeSpace) > MaxNode THEN
    FOR NodePtr = WorkPtr TO (WorkPtr + NodeBlockSize) - 1
       REM MarkNode WorkPtr		' this is more correct, but slower
       NodeSpace(NodePtr).fLink = unused	' faster than line above
    NEXT NodePtr
    NodePtr = MaxNode + 1
    MaxNode = MaxNode + NodeBlockSize
  ELSE
    ' couldn't grow the node allocation space, what do I do???
  END IF
END IF
END SUB

SUB MarkNode (NodePtr AS INTEGER)
' MarkNode -  Mark a list node as reuasable
NodeSpace(NodePtr).fLink = unused
END SUB


SUB InsertInPlace (ListHandle AS INTEGER, NodePtr AS INTEGER)
' InsertInPlace - Insert the node pointed to by NodePtr
'     into its proper position in the list
'     specified by ListHandle.
DIM Previous AS INTEGER, Current AS INTEGER, Found AS INTEGER
Previous = nil
Current = ListHeads(ListHandle)
Found = FALSE
DO WHILE (Current <> nil) AND NOT Found
  IF NodeSpace(NodePtr).fData.fKey.fValue _
   < NodeSpace(Current).fData.fKey.fValue THEN
    Found = TRUE
  ELSE
    Previous = Current
    Current = NodeSpace(Current).fLink
  END IF
LOOP
IF Previous = nil THEN
  PushNode ListHandle, NodePtr
ELSE
  InsertAfter Previous, NodePtr
END IF
END SUB

SUB InsertAfter (ListPtr AS INTEGER, NodePtr AS INTEGER)
' InsertAfter - Insert the node pointed to by NodePtr in
'   the list after the node pointed to by ListPtr
NodeSpace(NodePtr).fLink = NodeSpace(ListPtr).fLink
NodeSpace(ListPtr).fLink = NodePtr
END SUB

SUB DeleteAfter (ListPtr AS INTEGER)
' DeleteAfter - Delete the node pointed to by the node
'   pointed to by ListPtr
DIM aPtr AS INTEGER
aPtr = NodeSpace(ListPtr).fLink
NodeSpace(ListPtr).fLink = NodeSpace(aPtr).fLink
MarkNode aPtr
END SUB

SUB PushNode (ListHandle AS INTEGER, NodePtr AS INTEGER)
' PushNode -  Insert the node pointed to by NodePtr at
'               the head of the specified list.
NodeSpace(NodePtr).fLink = ListHeads(ListHandle)
ListHeads(ListHandle) = NodePtr
END SUB

SUB PopNode (ListHandle AS INTEGER)
' PopNode - remove the node at the head of the specified list
'   and mark it as reusable.
DIM WorkPtr AS INTEGER
WorkPtr = ListHeads(ListHandle)
ListHeads(ListHandle) = NodeSpace(WorkPtr).fLink
MarkNode WorkPtr
END SUB

SUB Traverse (ListHandle AS INTEGER, NodePtr AS INTEGER, ActionKey AS INTEGER)
' Traverse -  Return a pointer to the next node in the list.
'   If you specify the bit key value PeekNode as part of ActionKey,
'   Traverse will make a copy of the node and return a pointer to
'   the copy. When you do this, you should mark the node for later
'   reusal by calling MarkNode when you no longer need it.
'   ActionKey:
'     Use FindFirst for first node in the list,
'     use FindNext for subsequent calls
IF (ActionKey AND FindFirst) <> 0 THEN
  NodePtr = ListHeads(ListHandle)
ELSE
  IF NodePtr <> nil THEN
    NodePtr = NodeSpace(NodePtr).fLink
  END IF
END IF
END SUB

FUNCTION NextNode% (NodePtr AS INTEGER)
' NextNode% - 	Return the contents of the link field of the node\
'		pointed to by NodePtr.
NextNode% = NodeSpace(NodePtr).fLink
END FUNCTION

FUNCTION PreviousNode% (ListHandle AS INTEGER, NodePtr AS INTEGER)
' PreviousNode%	- Return a pointer to the node in the list
'		that precede the node pointed to by NodePtr.
'		Returns nil if node is first (it has no predecessor)
'		or if the node is not found.
DIM WorkPtr AS INTEGER
DIM Found AS INTEGER
WorkPtr = ListHeads(ListHandle)
DO WHILE (WorkPtr <> nil) AND NOT Found
  'IF NodeSpace(NodeSpace(WorkPtr).fLink) = NodeSpace(NodePtr) THEN
  IF NodeSpace(WorkPtr).fLink = NodePtr THEN
    Found = TRUE
  ELSE
    WorkPtr = NodeSpace(WorkPtr).fLink
  END IF
LOOP
PreviousNode% = WorkPtr
END FUNCTION

SUB StoreNodeData (NodePtr AS INTEGER, NodeData AS NodeDataType)
' StoreNodeData - Initialize the node pointed to by NodePtr to contain
'		  the data passed by the caller.
NodeSpace(NodePtr).fData = NodeData
END SUB

SUB GetNodeData (NodePtr AS INTEGER, NodeData AS NodeDataType)
' GetNodeData - Return in the argument NodeData the data record 
'		for the node pointed to by NodePtr 
NodeData = NodeSpace(NodePtr).fData
END SUB
